{% extends "layout.html" %}
{% set title = 'Sardana documentation' %}
{% block body %}

<h1>Welcome to Sardana documentation!</h1>

<p>
Sardana is the control program initially developed at <a class="reference external" href="http://www.cells.es/">ALBA</a>.
Our mission statement:
</p>

<blockquote><div><cite>
  Produce a modular, high performance, robust, and generic user environment
  for control applications in large and small installations. Make Sardana
  the generic user environment distributed in the TANGO project and the
  standard basis of collaborations in control.
</cite></div></blockquote>

<script type="text/javascript" src="_static/slideshow.js"></script>

<div id="gallery">
  <img src="_static/gallery/gallery01.png" />
  <img src="_static/gallery/gallery02.png" />
</div>

<!--
<div class="figure align-center">
<img alt="Sardana" src="_static/sardana_screenshot.png" style="width: 70%;" />
</div>
-->

<p>
  An <a class="reference internal" href="users/overview.html#sardana-overview"><em>Overview</em></a>
  guide will help you getting started with the basic sardana concepts. 
  The <a class="reference internal" href="users/faq.html#sardana-faq"><em>FAQ</em></a>
  will answer many of your questions.
</p>

<p>
  For sampling, see the <a class="reference internal" href="users/screenshots.html#screenshots"><em>Screenshots</em></a>
  and <a class="reference internal" href="devel/examples/examples.html#devel-examples"><em>Examples</em></a>
  chapters.
</p>

<!--
<p>
    A PDF version can be downloaded from <a href="sardana.pdf">here</a>.
</p>
-->

<p>
    For convenience here are the links to other versions:<br/>
    <a href="http://www.sardana-controls.org/development/">Development</a>
    &nbsp;
    <a href="http://www.sardana-controls.org/">Latest stable</a>
    &nbsp;
    <!-- <a href="http://www.sardana-controls.org/v100/doc/html/">1.0.0</a> -->
</p>

{% endblock %}
