<h3>Download</h3>
<p>Current version: <b>{{ release }}</b></p>
<p>Get sardana from the <a href="http://pypi.python.org/pypi/sardana">PyPi</a>,
or install it with:</p>
<pre>easy_install -U sardana</pre>

<!--
<h3>PDF</h3>
<p>A PDF version <a href="sardana.pdf">here</a>.
-->

<h3>Development</h3>
<p>Latest <a href="http://www.sardana-controls.org/development/">development version docs</a>
are also available.</p>
